# pylint: disable=invalid-name
# pylint: disable=bare-except

"""
Useful routines for symbolic manipulation of Jones vectors and matrices

Todo:
    * tests and documentation

Scott Prahl
Apr 2019
"""

import sympy
import pypolar.sym_fresnel as sym_fresnel

__all__ = ['op_linear_polarizer',
           'op_retarder',
           'op_attenuator',
           'op_mirror',
           'op_rotation',
           'op_quarter_wave_plate',
           'op_half_wave_plate',
           'op_fresnel_reflection',
           'op_fresnel_transmission',
           'field_linear',
           'field_left_circular',
           'field_right_circular',
           'field_horizontal',
           'field_vertical']


def op_linear_polarizer(theta):
    """
    Jones matrix operator for a linear polarizer rotated about a normal to
    the surface of the polarizer.
    theta: rotation angle measured from the horizontal plane [radians]
    """

    return sympy.Matrix([[sympy.cos(theta)**2, sympy.sin(theta) * sympy.cos(theta)],
                         [sympy.sin(theta) * sympy.cos(theta), sympy.sin(theta)**2]])


def op_retarder(theta, delta):
    """
    Jones matrix operator for an optical retarder rotated about a normal to
    the surface of the retarder.
    theta: rotation angle between fast-axis and the horizontal plane [radians]
    delta: phase delay introduced between fast and slow-axes         [radians]
    """
    P = sympy.exp(+delta / 2 * sympy.numbers.I)
    Q = sympy.exp(-delta / 2 * sympy.numbers.I)
    D = sympy.sin(delta / 2) * 2 * sympy.numbers.I
    C = sympy.cos(theta)
    S = sympy.sin(theta)
    return sympy.Matrix([[C * C * P + S * S * Q, C * S * D],
                         [C * S * D, C * C * Q + S * S * P]])


def op_attenuator(t):
    """
    Jones matrix operator for an optical attenuator.
    t: fraction of light passing through attenuator [---]
    """

    return sympy.Matrix([[t/ 2, 0], [0, t/ 2]])


def op_neutral_density_filter(nd):
    """
    Jones matrix operator for a neutral density filter with decadic attenuation
    nd: base ten optical density  [---]
    """

    return sympy.Matrix([[10**-nd / 2, 0], [0, 10**-nd / 2]])

def op_mirror():
    """
    Jones matrix operator for a perfect mirror.
    """
    return sympy.Matrix([[1, 0], [0, -1]])


def op_rotation(theta):
    """
    Jones matrix operator to rotate light about the optical axis.

    Args:
        theta : angle of rotation about optical axis  [radians]
    Returns:
        2x2 matrix of the rotation operator           [-]
    """
    return sympy.Matrix([[sympy.cos(theta), sympy.sin(theta)],
                     [-sympy.sin(theta), sympy.cos(theta)]])


def op_quarter_wave_plate(theta):
    """
    Jones matrix operator for an quarter-wave plate rotated about a normal to
    the surface of the plate.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the quarter-wave plate operator     [-]
    """

    return op_retarder(theta, sympy.numbers.pi / 2)


def op_half_wave_plate(theta):
    """
    Jones matrix operator for an half-wave plate rotated about a normal to
    the surface of the plate.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the half-wave plate operator     [-]
    """

    return op_retarder(theta, sympy.numbers.pi)


def op_fresnel_reflection(m, theta):
    """
    Jones matrix operator for Fresnel reflection at angle theta
    Args:
        m :     complex index of refraction   [-]
        theta : angle from normal to surface  [radians]
    Returns:
        2x2 matrix of the Fresnel transmission operator     [-]
    """
    return sympy.Matrix([[sym_fresnel.r_par(m, theta), 0],
                     [0, sym_fresnel.r_per(m, theta)]])


def op_fresnel_transmission(m, theta):
    """
    Jones matrix operator for Fresnel transmission at angle theta

    *** THIS IS ALMOST CERTAINLY WRONG ***
    Args:
        m :     complex index of refraction       [-]
        theta : angle from normal to surface      [radians]
    Returns:
        2x2 Fresnel transmission operator           [-]
    """
    c = sympy.cos(theta)
    d = sympy.sqrt(m * m - sympy.sin(theta)**2, dtype=sympy.complex)
    if m.imag == 0:
        d = sympy.conjugate(d)
    a = sympy.sqrt(d/c)
    return a*sympy.Matrix([[sym_fresnel.t_par(m, theta), 0], [0, sym_fresnel.t_per(m, theta)]])


def field_linear(theta):
    """Jones vector for linear polarized light at angle theta from horizontal plane"""

    return sympy.Matrix([sympy.cos(theta), sympy.sin(theta)])


def field_right_circular():
    """Jones Vector corresponding to right circular polarized light"""

    return 1 / sympy.sqrt(2) * sympy.Matrix([1, -sympy.numbers.I])


def field_left_circular():
    """Jones Vector corresponding to left circular polarized light"""

    return 1 / sympy.sqrt(2) * sympy.Matrix([1, sympy.numbers.I])


def field_horizontal():
    """Jones Vector corresponding to horizontal polarized light"""

    return field_linear(0)


def field_vertical():
    """Jones Vector corresponding to vertical polarized light"""

    return field_linear(sympy.numbers.pi / 2)


def intensity(J):
    """
    Returns the intensity
    """
    return sympy.abs(J[0])**2 + sympy.abs(J[1])**2


def phase(J):
    """
    Returns the phase
    """
    gamma = sympy.arg(J[1]) - sympy.arg(J[0])
    return gamma


def ellipse_orientation(J):
    """
    Returns the angle between the major semi-axis and the x-axis of
    the polarization ellipse (sometimes called the azimuth or psi)
    """
    Exo, Eyo = sympy.abs(J)
    delta = phase(J)
    numer = 2 * Exo * Eyo * sympy.cos(delta)
    denom = Exo**2 - Eyo**2
    psi = 0.5 * sympy.arctan2(numer, denom)
    return psi


def ellipse_ellipticity(J):
    """
    Returns the ellipticty of the polarization ellipse.
    """
    delta = phase(J)
    psi = ellipse_orientation(J)
    chi = 0.5 * sympy.arcsin(sympy.sin(2 * psi) * sympy.sin(delta))
    return chi


def ellipse_axes(J):
    """
    Returns the semi-major and semi-minor axes of the polarization ellipse.
    """
    Exo, Eyo = sympy.abs(J)
    psi = ellipse_orientation(J)
    delta = phase(J)
    C = sympy.cos(psi)
    S = sympy.sin(psi)
    asqr = (Exo * C)**2 + (Eyo * S)**2 + 2 * Exo * Eyo * C * S * sympy.cos(delta)
    bsqr = (Exo * S)**2 + (Eyo * C)**2 - 2 * Exo * Eyo * C * S * sympy.cos(delta)
    return sympy.sqrt(abs(asqr)), sympy.sqrt(abs(bsqr))




# pylint: disable=invalid-name
# pylint: disable=bare-except

"""
Useful basic routines for visualizing polarization

Todo:
    * improve documentation and testing

Scott Prahl
Mar 2019
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.animation as animation
import pypolar.fresnel
import pypolar.mueller
import pypolar.jones

from mpl_toolkits.mplot3d import Axes3D

#from IPython.display import HTML
#import mpl_toolkits.mplot3d.axes3d as axes3d


__all__ = ['draw_jones_field',
           'draw_jones_animated',
           'draw_jones_ellipse',
           'draw_stokes_ellipse',
           'draw_stokes_field',
           'draw_stokes_animated']


def _draw_optical_axis_3d(J, ax, last=4 * np.pi):
    """
    Draw the optical axis in a 3D plot

    Args:
        J:    Jones vector
        ax:   matplotlib axis to use
        last: length of optical axis
    """
    h_amp, v_amp = abs(J)
    the_max = max(h_amp, v_amp) * 1.1

    ax.plot([0, last], [0, 0], [0, 0], 'k')
    ax.plot([0, 0], [-the_max, the_max], [0, 0], 'g')
    ax.plot([0, 0], [0, 0], [-the_max, the_max], 'b')


def _draw_h_field_3d(J, ax, offset, last=4 * np.pi):
    """
    Draw the horizontal electric field in a 3D plot

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
        last:   length of optical axis
    """
    t = np.linspace(0, last, 100)
    x = t
    y = np.abs(J[0]) * np.cos(t + offset - np.angle(J[0]))
    z = 0
    ax.plot(x, y, z, ':g')


def _draw_v_field_3d(J, ax, offset, last=4 * np.pi):
    """
    Draw the vertical electric field in a 3D plot

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
        last:   length of optical axis
    """
    t = np.linspace(0, last, 100)
    x = t
    y = 0 * t
    z = np.abs(J[1]) * np.cos(t + offset - np.angle(J[1]))
    ax.plot(x, y, z, ':b')


def _draw_total_field_3d(J, ax, offset, last=4 * np.pi):
    """
    Draw the total electric field in a 3D plot

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
        last:   length of optical axis
    """
    t = np.linspace(0, last, 100)
    x = t
    y = np.abs(J[0]) * np.cos(t + offset - np.angle(J[0]))
    z = np.abs(J[1]) * np.cos(t + offset - np.angle(J[1]))
    ax.plot(x, y, z, 'r')


def _draw_projected_vector_3d(J, ax, offset):
    """
    Draw the projection vector of the polarization field in 3D

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
    """

    y = np.abs(J[0]) * np.cos(offset - np.angle(J[0]))
    z = np.abs(J[1]) * np.cos(offset - np.angle(J[1]))

    x1, y1, z1 = 0, y, 0
    x2, y2, z2 = 0, y, z
    ax.plot([x1, x2], [y1, y2], [z1, z2], 'g--')

    x1, y1, z1 = 0, 0, z
    ax.plot([x1, x2], [y1, y2], [z1, z2], 'b--')

    x1, y1, z1 = 0, 0, 0
    ax.plot([x1, x2], [y1, y2], [z1, z2], 'r')
    ax.plot([x2], [y2], [z2], 'ro')


def _draw_3D_field(J, ax, offset):
    """
    Draw a representation of the polarization fields in 3D

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
    """
    _draw_optical_axis_3d(J, ax, offset)
    _draw_h_field_3d(J, ax, offset)
    _draw_v_field_3d(J, ax, offset)
    _draw_total_field_3d(J, ax, offset)
    _draw_projected_vector_3d(J, ax, offset)

    ax.grid(False)
    ax.axis('off')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])


def _draw_2D_field(J, ax, offset):
    """
    Draw a simple 2D representation of the projected field

    Args:
        J:      Jones vector
        ax:     matplotlib axis to use
        offset: starting point
    """
    h_amp, v_amp = np.abs(J)
    h_phi, v_phi = np.angle(J)
    the_max = max(h_amp, v_amp) * 1.1

    plt.plot([-the_max, the_max], [0, 0], 'g')
    plt.plot([0, 0], [-the_max, the_max], 'b')

    t = np.linspace(0, 2 * np.pi, 100)
    x = h_amp * np.cos(t + offset - h_phi)
    y = v_amp * np.cos(t + offset - v_phi)
    ax.plot(x, y, 'k')

    x = h_amp * np.cos(offset - h_phi)
    y = v_amp * np.cos(offset - v_phi)
    ax.plot(x, y, 'ro')
    ax.plot([x, x], [0, y], 'g--')
    ax.plot([0, x], [y, y], 'b--')
    ax.plot([0, x], [0, y], 'r')

    ax.set_xlim(-the_max, the_max)
    ax.set_ylim(-the_max, the_max)
    ax.set_aspect('equal')
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])


def draw_jones_ellipse(J):
    """
    Draw a simple 2D representation of the projected field

    Args:
        J:      Jones vector
    """
    Exo, Eyo = np.abs(J)
    phix, phiy = np.angle(J)
    the_max = max(Exo, Eyo) * 1.2

    the_max = the_max
    the_max = the_max
    plt.axes().set_aspect('equal')
    plt.plot([-the_max, the_max], [0, 0], 'k')
    plt.plot([0, 0], [-the_max, the_max], 'k')

    plt.plot([-Exo, -Exo, Exo, Exo, -Exo], [-Eyo, Eyo, Eyo, -Eyo, -Eyo], ':g')
    plt.annotate(r' $E_{x0}$', xy=(Exo, 0), va='bottom', ha='left')
    plt.annotate(r'$-E_{x0} $', xy=(-Exo, 0), va='bottom', ha='right')
    plt.annotate(r'$E_{y0}$', xy=(0, Eyo), va='bottom', ha='left')
    plt.annotate(r'$-E_{y0}$', xy=(0, -Eyo), va='top', ha='left')
    plt.annotate(r'  $\psi$', xy=(0, 0), va='bottom', ha='left')

    t = np.linspace(0, 2 * np.pi, 100)
    plt.plot(Exo * np.cos(t + phix), Eyo * np.cos(t + phiy), 'b')

    psi = pypolar.jones.ellipse_orientation(J)
    M = np.sqrt(pypolar.jones.intensity(J))
    plt.plot([-M * np.cos(psi), M * np.cos(psi)],
             [-M * np.sin(psi), M * np.sin(psi)])
    plt.xlim(-the_max, the_max)
    plt.ylim(-the_max, the_max)
    plt.xticks([])
    plt.yticks([])
    return plt


def draw_stokes_ellipse(S):
    """
    Draw a simple 2D representation of the projected field

    Args:
        S:      Stokes vector
    """
    Exo = np.sqrt((S[0]+S[1])/2)
    Eyo = np.sqrt((S[0]-S[1])/2)
    phix = 0
    phiy = np.arcsin(S[2]/(2*Exo*Eyo))
    the_max = max(Exo, Eyo) * 1.2

    the_max = the_max
    the_max = the_max
    plt.axes().set_aspect('equal')
    plt.plot([-the_max, the_max], [0, 0], 'k')
    plt.plot([0, 0], [-the_max, the_max], 'k')

    plt.plot([-Exo, -Exo, Exo, Exo, -Exo], [-Eyo, Eyo, Eyo, -Eyo, -Eyo], ':g')
    plt.annotate(r' $E_{x0}$', xy=(Exo, 0), va='bottom', ha='left')
    plt.annotate(r'$-E_{x0} $', xy=(-Exo, 0), va='bottom', ha='right')
    plt.annotate(r'$E_{y0}$', xy=(0, Eyo), va='bottom', ha='left')
    plt.annotate(r'$-E_{y0}$', xy=(0, -Eyo), va='top', ha='left')
    plt.annotate(r'  $\psi$', xy=(0, 0), va='bottom', ha='left')

    t = np.linspace(0, 2 * np.pi, 100)
    plt.plot(Exo * np.cos(t + phix), Eyo * np.cos(t + phiy), 'b')

    psi = pypolar.mueller.ellipse_orientation(S)
    M = np.sqrt(pypolar.mueller.intensity(S))
    plt.plot([-M * np.cos(psi), M * np.cos(psi)],
             [-M * np.sin(psi), M * np.sin(psi)])
    plt.xlim(-the_max, the_max)
    plt.ylim(-the_max, the_max)
    plt.xticks([])
    plt.yticks([])
    return plt


def draw_jones_field(J, offset=0):
    """
    Draw 3D and 2D representations of the polarization field

    Args:
        J:      Jones vector
        offset: starting point
    """
    plt.figure(figsize=(8, 4))
    gs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])

    ax1 = plt.subplot(gs[0], projection='3d')
    _draw_3D_field(J, ax1, offset)

    ax2 = plt.subplot(gs[1])
    _draw_2D_field(J, ax2, offset)
    return plt


def _animation_update(offset, J, ax1, ax2):
    """
    function to draw the next animation frame

    Args:
        offset: starting phase for drawings
        J:      Jones vector
        ax1:    matplotlib axis for 3D plot
        ax2:    matplotlib axis for 2D plot
    """
    ax1.clear()
    ax2.clear()
    _draw_3D_field(J, ax1, offset)
    _draw_2D_field(J, ax2, offset)
    return ax1, ax2


def draw_jones_animated(J):
    """
    Animate 3D and 2D representations of the polarization field

    Args:
        J:      Jones vector
    """
    fig = plt.figure(figsize=(8, 4))
    gs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])

    ax1 = plt.subplot(gs[0], projection='3d')
    _draw_3D_field(J, ax1, 0)

    ax2 = plt.subplot(gs[1])
    _draw_2D_field(J, ax2, 0)

    ani = animation.FuncAnimation(fig, _animation_update,
                                  frames=np.linspace(0, 2 * np.pi, 64),
                                  fargs=(J, ax1, ax2))
    return ani


def draw_stokes_ellipse(S):
    """
    Draw a 2D representation of the polarization state of S

    Args:
        S:      Stokes vector
        offset: starting point
    Returns:
        a matplotlib object with the graph
    """
    J = pypolar.mueller.stokes_to_jones(S)
    aplt = draw_jones_ellipse(J)
    return aplt


def draw_stokes_field(S, offset=0):
    """
    Draw a 2D and 3D representation of the polarization

    Args:
        S:      Stokes vector
        offset: starting point
    """
    J = pypolar.mueller.stokes_to_jones(S)
    aplt = draw_jones_field(J, offset)
    return aplt


def draw_stokes_animated(S):
    """
    Draw animated 2D and 3D representations of the polarization

    Args:
        S:      Stokes vector
    """
    J = pypolar.mueller.stokes_to_jones(S)
    ani = draw_jones_animated(J)
    return ani



# pylint: disable=invalid-name
"""
Useful basic routines for managing Fresnel reflection

To do:
    * Make sure routines work for arrays of m or of theta_i
    * add ellipsometry routines for one layer

Scott Prahl
Apr 2018
"""

import numpy as np

__all__ = ['r_par',
           'r_per',
           't_par',
           't_per',
           'R_par',
           'R_per',
           'T_par',
           'T_per',
           'R_unpolarized',
           'T_unpolarized',
           'ellipsometry_rho',
           'ellipsometry_index',
           'ellipsometry_parameters']


def r_par(m, theta_i):
    """
    Calculates the reflected amplitude for parallel polarized light

    Args:
        m :       complex index of refraction   [-]
        theta_i : angle from normal to surface  [radians]
    Returns:
        reflected fraction of parallel field    [-]
    """
    c = m * m * np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m * m - s * s, dtype=np.complex) # m*cos(theta_t)
    if m.imag == 0:
        d = np.conjugate(d)
    rp = (c - d) / (c + d)
    return np.real_if_close(rp)


def r_per(m, theta_i):
    """
    Calculates the reflected amplitude for perpendicular polarized light

    Args:
        m :       complex index of refraction     [-]
        theta_i : incidence angle from normal     [radians]
    Returns:
        reflected fraction of perpendicular field [-]
    """
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m * m - s * s, dtype=np.complex) # m*cos(theta_t)
    if m.imag == 0:
        d = np.conjugate(d)
    rs = (c - d) / (c + d)
    return np.real_if_close(rs)


def t_par(m, theta_i):
    """
    Calculates the transmitted amplitude for parallel polarized light

    Args:
        m :       complex index of refraction  [-]
        theta_i : incidence angle from normal  [radians]
    Returns:
        transmitted fraction of parallel field [-]
    """
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m * m - s * s, dtype=np.complex) # m*cos(theta_t)
    if m.imag == 0:
        d = np.conjugate(d)
    tp = 2 * c * m / (m * m * c + d)
    return np.real_if_close(tp)


def t_per(m, theta_i):
    """
    Calculates the transmitted amplitude for perpendicular polarized light

    Args:
        m :     complex index of refraction         [-]
        theta_i : incidence angle from normal       [radians]
    Returns:
        transmitted fraction of perpendicular field [-]
    """
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m * m - s * s, dtype=np.complex) # m*cos(theta_t)
    if m.imag == 0:
        d = np.conjugate(d)
    ts = 2 * c / (c + d)
    return np.real_if_close(ts)


def R_par(m, theta_i):
    """
    Fraction of parallel-polarized light that is reflected (R_p).

    Calculates reflected fraction of incident power (or flux) assuming that
    the E-field of the incident light is parallel to the plane of incidence

    Args:
        m :       complex index of refraction [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected power                       [-]
    """
    return abs(r_par(m, theta_i))**2


def R_per(m, theta_i):
    """
    Fraction of perpendicular-polarized light that is reflected (R_s).

    Calculates reflected fraction of incident power (or flux) assuming that
    the E-field of the incident light is perpendicular to the plane of incidence

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return abs(r_per(m, theta_i))**2


def T_par(m, theta_i):
    """
    Fraction of parallel-polarized light that is transmitted (T_p).

    Calculates transmitted fraction of incident power (or flux) assuming that
    the E-field of the incident light is parallel to the plane of incidence

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        transmitted irradiance                [-]
    """
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m * m - s * s, dtype=np.complex) # m*cos(theta_t)
    tp = 2 * c * m / (m * m * c + d)
    return np.real(d / c * abs(tp)**2)


def T_per(m, theta_i):
    """
    Fraction of perpendicular-polarized light that is transmitted (T_s).

    Calculates transmitted fraction of incident power (or flux) assuming that
    the E-field of the incident light is perpendicular to the plane of incidence

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        transmitted field amplitude           [-]
    """
    c = np.cos(theta_i)
    s = np.sin(theta_i)
    d = np.sqrt(m * m - s * s, dtype=np.complex) # m*cos(theta_t)
    ts = 2 * c / (c + d)
    return np.real(d / c * abs(ts)**2)


def R_unpolarized(m, theta_i):
    """
    Fraction of unpolarized light that is reflected.

    Calculates reflection fraction of incident power (or flux) assuming that
    the incident light is unpolarized

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return (R_par(m, theta_i) + R_per(m, theta_i)) / 2


def T_unpolarized(m, theta_i):
    """
    Fraction of unpolarized light that is transmitted.

    Calculates transmitted fraction of incident power (or flux) assuming that
    the incident light is unpolarized

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        reflected irradiance                  [-]
    """
    return (T_par(m, theta_i) + T_per(m, theta_i)) / 2


def ellipsometry_rho(m, theta_i):
    """
    Calculate the ellipsometer parameter rho

    Args:
        m :     complex index of refraction   [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        ellipsometer parameter rho            [-]
    """
    return r_par(m, theta_i) / r_per(m, theta_i)


def ellipsometry_index(rho, theta_i):
    """
    Calculate the index of refraction for an isotropic sample

    Args:
        rho :  r_par/r_per                    [-]
        theta_i : incidence angle from normal [radians]
    Returns:
        complex index of refraction           [-]
    """
    e_index = np.sqrt(1 - 4 * rho * np.sin(theta_i)**2 / (1 + rho)**2)
    return np.tan(theta_i) * e_index


def ellipsometry_parameters(phi, signal, P):
    """
    Recover ellipsometer parameters Delta and tan(psi) by fitting to

             I_DC + I_S*sin(2*phi)+I_C*cos(2*phi)

    Args:
        phi    - array of analyzer angles               [radians]
        signal - array of ellipsometer intensities      [AU]
        P      - incident polarization azimuthal angle  [radians]
    """

    I_DC = np.average(signal)
    I_S = 2 * np.average(signal * np.sin(2 * phi))
    I_C = 2 * np.average(signal * np.cos(2 * phi))

    tanP = np.tan(P)
    arg = I_S / np.sqrt(abs(I_DC**2 - I_C**2)) * np.sign(tanP)
    if arg > 1:
        Delta = 0
    elif arg < -1:
        Delta = np.pi
    else:
        Delta = np.arccos(arg)

    tanpsi = np.sqrt(abs(I_DC + I_C) / abs(I_DC - I_C)) * np.abs(tanP)

    return Delta, tanpsi



# pylint: disable=invalid-name
# pylint: disable=bare-except

"""
Useful basic routines for managing polarization using the Jones calculus

Todo:
    * improve documentation of each routine
    * modify interpret() when phase difference differs by more than 2pi
    * improve interpret() to give angle for elliptical polarization

Scott Prahl
Apr 2018
"""

import numpy as np
import pypolar.fresnel

__all__ = ['op_linear_polarizer',
           'op_retarder',
           'op_attenuator',
           'op_mirror',
           'op_rotation',
           'op_quarter_wave_plate',
           'op_half_wave_plate',
           'op_fresnel_reflection',
           'op_fresnel_transmission',
           'field_linear',
           'field_left_circular',
           'field_right_circular',
           'field_horizontal',
           'field_vertical',
           'interpret',
           'intensity',
           'phase',
           'ellipse_azimuth',
           'ellipse_ellipticity',
           'ellipse_orientation',
           'ellipse_axes',
           'poincare_point',
           'jones_op_to_mueller_op']


def op_linear_polarizer(theta):
    """
    Jones matrix operator for a linear polarizer rotated about a normal to
    the surface of the polarizer.
    theta: rotation angle measured from the horizontal plane [radians]
    """

    return np.matrix([[np.cos(theta)**2, np.sin(theta) * np.cos(theta)],
                      [np.sin(theta) * np.cos(theta), np.sin(theta)**2]])


def op_retarder(theta, delta):
    """
    Jones matrix operator for an optical retarder rotated about a normal to
    the surface of the retarder.
    theta: rotation angle between fast-axis and the horizontal plane [radians]
    delta: phase delay introduced between fast and slow-axes         [radians]
    """
    P = np.exp(+delta / 2 * 1j)
    Q = np.exp(-delta / 2 * 1j)
    D = np.sin(delta / 2) * 2j
    C = np.cos(theta)
    S = np.sin(theta)
    return np.array([[C * C * P + S * S * Q, C * S * D],
                     [C * S * D, C * C * Q + S * S * P]])


def op_attenuator(t):
    """
    Jones matrix operator for an optical attenuator.
    f: fraction of light getting through attenuator  [---]
    """
    return np.matrix([[t / 2, 0], [0, t / 2]])


def op_mirror():
    """
    Jones matrix operator for a perfect mirror.
    """
    return np.matrix([[1, 0], [0, -1]])


def op_rotation(theta):
    """
    Jones matrix operator to rotate light about the optical axis.

    Args:
        theta : angle of rotation about optical axis  [radians]
    Returns:
        2x2 matrix of the rotation operator           [-]
    """
    return np.array([[np.cos(theta), np.sin(theta)],
                     [-np.sin(theta), np.cos(theta)]])


def op_quarter_wave_plate(theta):
    """
    Jones matrix operator for an quarter-wave plate rotated about a normal to
    the surface of the plate.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the quarter-wave plate operator     [-]
    """

    return op_retarder(theta, np.pi / 2)


def op_half_wave_plate(theta):
    """
    Jones matrix operator for an half-wave plate rotated about a normal to
    the surface of the plate.

    Args:
        theta : angle from fast-axis to horizontal plane  [radians]
    Returns:
        2x2 matrix of the half-wave plate operator     [-]
    """

    return op_retarder(theta, np.pi)


def op_fresnel_reflection(m, theta):
    """
    Jones matrix operator for Fresnel reflection at angle theta
    Args:
        m :     complex index of refraction   [-]
        theta : angle from normal to surface  [radians]
    Returns:
        2x2 matrix of the Fresnel transmission operator     [-]
    """
    return np.array([[pypolar.fresnel.r_par(m, theta), 0],
                     [0, pypolar.fresnel.r_per(m, theta)]])


def op_fresnel_transmission(m, theta):
    """
    Jones matrix operator for Fresnel transmission at angle theta

    *** THIS IS ALMOST CERTAINLY WRONG ***
    Args:
        m :     complex index of refraction       [-]
        theta : angle from normal to surface      [radians]
    Returns:
        2x2 Fresnel transmission operator           [-]
    """
    c = np.cos(theta)
    d = np.sqrt(m * m - np.sin(theta)**2, dtype=np.complex)
    if m.imag == 0:
        d = np.conjugate(d)
    a = np.sqrt(d/c)
    return a*np.array([[pypolar.fresnel.t_par(m, theta), 0], [0, pypolar.fresnel.t_per(m, theta)]])


def field_linear(theta):
    """Jones vector for linear polarized light at angle theta from horizontal plane"""

    return np.array([np.cos(theta), np.sin(theta)])


def field_right_circular():
    """Jones Vector corresponding to right circular polarized light"""

    return 1 / np.sqrt(2) * np.array([1, -1j])


def field_left_circular():
    """Jones Vector corresponding to left circular polarized light"""

    return 1 / np.sqrt(2) * np.array([1, 1j])


def field_horizontal():
    """Jones Vector corresponding to horizontal polarized light"""

    return field_linear(0)


def field_vertical():
    """Jones Vector corresponding to vertical polarized light"""

    return field_linear(np.pi / 2)


def interpret(J):
    '''
    Interprets a Jones vector (Original version by Alexander Miles 2013)

    Parameters
    J     : A Jones vector, may be complex

    Examples
    -------
    interpret([1, -1j]) --> "Right circular polarization"

    interpret([0.5, 0.5]) -->
                      "Linear polarization at 45.000000 degrees CCW from x-axis"

    interpret( np.array([exp(-1j*pi), exp(-1j*pi/3)]) ) -->
                "Left elliptical polarization, rotated with respect to the axes"
    '''

    try:
        j1, j2 = J
    except:
        print("Jones vector must have two elements")
        return 0

    eps = 1e-12
    mag1, p1 = abs(j1), np.angle(j1)
    mag2, p2 = abs(j2), np.angle(j2)

    if np.remainder(p1 - p2, np.pi) < eps:
        ang = np.arctan2(mag2, mag1) * 180 / np.pi
        return "Linear polarization at %f degrees CCW from x-axis" % ang

    if abs(mag1 - mag2) < eps:
        if abs(p1 - p2 - np.pi / 2) < eps:
            s = "Right circular polarization"
        elif p1 > p2:
            s = "Right elliptical polarization, rotated with respect to the axes"
        if (p1 - p2 + np.pi / 2) < eps:
            s = "Left circular polarization"
        elif p1 < p2:
            s = "Left elliptical polarization, rotated with respect to the axes"
    else:
        if p1 - p2 == np.pi / 2:
            s = "Right elliptical polarization, non-rotated"
        elif p1 > p2:
            s = "Right elliptical polarization, rotated with respect to the axes"
        if p1 - p2 == -np.pi / 2:
            s = "Left circular polarization, non-rotated"
        elif p1 < p2:
            s = "Left elliptical polarization, rotated with respect to the axes"
    return s


def normalize_vector(J):
    """
    Normalizes a vector by dividing each part by common number.
    After normalization the magnitude should be equal to ~1.
    """
    norm = np.linalg.norm(J)
    if norm == 0:
        return J
    return J / norm


def intensity(J):
    """
    Returns the intensity
    """
    return np.abs(J[0])**2 + np.abs(J[1])**2


def phase(J):
    """
    Returns the phase
    """
    gamma = np.angle(J[1]) - np.angle(J[0])
    return gamma


def ellipse_orientation(J):
    """
    Returns the angle between the major semi-axis and the x-axis of
    the polarization ellipse (sometimes called the azimuth or psi)
    """
    Exo, Eyo = np.abs(J)
    delta = phase(J)
    numer = 2 * Exo * Eyo * np.cos(delta)
    denom = Exo**2 - Eyo**2
    psi = 0.5 * np.arctan2(numer, denom)
    return psi


def ellipse_ellipticity(J):
    """
    Returns the ellipticty of the polarization ellipse.
    """
    delta = phase(J)
    psi = ellipse_orientation(J)
    chi = 0.5 * np.arcsin(np.sin(2 * psi) * np.sin(delta))
    return chi


def ellipse_azimuth(J):
    """
    Returns the angle between the major semi-axis and the x-axis of
    the polarization ellipse.
    """
    Exo, Eyo = np.abs(J)
    alpha = np.arctan2(Eyo, Exo)
    return alpha


def ellipse_axes(J):
    """
    Returns the semi-major and semi-minor axes of the polarization ellipse.
    """
    Exo, Eyo = np.abs(J)
    psi = ellipse_orientation(J)
    delta = phase(J)
    C = np.cos(psi)
    S = np.sin(psi)
    asqr = (Exo * C)**2 + (Eyo * S)**2 + 2 * Exo * Eyo * C * S * np.cos(delta)
    bsqr = (Exo * S)**2 + (Eyo * C)**2 - 2 * Exo * Eyo * C * S * np.cos(delta)
    return np.sqrt(abs(asqr)), np.sqrt(abs(bsqr))


def poincare_point(J):
    """
    Returns the point the Poincaré sphere
    """
    longitude = 2 * ellipse_orientation(J)
    a, b = ellipse_axes(J)
    latitude = 2 * np.arctan2(b, a)
    return latitude, longitude


def jones_op_to_mueller_op(J):
    """
    Converts a complex 2x2 Jones matrix to a real 4x4 Mueller matrix

    Hauge, Muller, and Smith, "Conventions and Formulas for Using the Mueller-
    Stokes Calculus in Ellipsometry," Surface Science, 96, 81-107 (1980)
    Args:
        J:      Jones matrix
    Returns
        equivalent 4x4 Mueller matrix
    """
    M = np.zeros(shape=[4, 4], dtype=np.complex)
    C = np.conjugate(J)
    M[0, 0] = J[0, 0] * C[0, 0] + J[0, 1] * C[0, 1] + \
        J[1, 0] * C[1, 0] + J[1, 1] * C[1, 1]
    M[0, 1] = J[0, 0] * C[0, 0] + J[1, 0] * C[1, 0] - \
        J[0, 1] * C[0, 1] - J[1, 1] * C[1, 1]
    M[0, 2] = J[0, 1] * C[0, 0] + J[1, 1] * C[1, 0] + \
        J[0, 0] * C[0, 1] + J[1, 0] * C[1, 1]
    M[0, 3] = 1j * (J[0, 1] * C[0, 0] + J[1, 1] * C[1, 0] -
                    J[0, 0] * C[0, 1] - J[1, 0] * C[1, 1])
    M[1, 0] = J[0, 0] * C[0, 0] + J[0, 1] * C[0, 1] - \
        J[1, 0] * C[1, 0] - J[1, 1] * C[1, 1]
    M[1, 1] = J[0, 0] * C[0, 0] - J[1, 0] * C[1, 0] - \
        J[0, 1] * C[0, 1] + J[1, 1] * C[1, 1]
    M[1, 2] = J[0, 0] * C[0, 1] + J[0, 1] * C[0, 0] - \
        J[1, 0] * C[1, 1] - J[1, 1] * C[1, 0]
    M[1, 3] = 1j * (J[0, 1] * C[0, 0] + J[1, 0] * C[1, 1] -
                    J[1, 1] * C[1, 0] - J[0, 0] * C[0, 1])
    M[2, 0] = J[0, 0] * C[1, 0] + J[1, 0] * C[0, 0] + \
        J[0, 1] * C[1, 1] + J[1, 1] * C[0, 1]
    M[2, 1] = J[0, 0] * C[1, 0] + J[1, 0] * C[0, 0] - \
        J[0, 1] * C[1, 1] - J[1, 1] * C[0, 1]
    M[2, 2] = J[0, 0] * C[1, 1] + J[0, 1] * C[1, 0] + \
        J[1, 0] * C[0, 1] + J[1, 1] * C[0, 0]
    M[2, 3] = 1j * (-J[0, 0] * C[1, 1] + J[0, 1] * C[1, 0] -
                    J[1, 0] * C[0, 1] + J[1, 1] * C[0, 0])
    M[3, 0] = 1j * (J[0, 0] * C[1, 0] + J[0, 1] * C[1, 1] -
                    J[1, 0] * C[0, 0] - J[1, 1] * C[0, 1])
    M[3, 1] = 1j * (J[0, 0] * C[1, 0] - J[0, 1] * C[1, 1] -
                    J[1, 0] * C[0, 0] + J[1, 1] * C[0, 1])
    M[3, 2] = 1j * (J[0, 0] * C[1, 1] + J[0, 1] * C[1, 0] -
                    J[1, 0] * C[0, 1] - J[1, 1] * C[0, 0])
    M[3, 3] = J[0, 0] * C[1, 1] - J[0, 1] * C[1, 0] - \
        J[1, 0] * C[0, 1] + J[1, 1] * C[0, 0]
    MM = M.real / 2
    return MM



"""
================================================================================
pypolar: Analysis of polarization using the Jones and/or the Mueller calculus
================================================================================

    http://github.com/scottprahl/pypolar

Usage:

import pypolar.jones as jones
import pypolar.mueller as mueller

light = jones.field_horizontal()
print("Jones vector for horizontally-polarized light")
print(light)

light = mueller.field_left_circular()
print("Stokes vector for left circularly polarized light")
print(light)


"""

__author__ = 'Scott Prahl'
__version__ = '0.5.0'



# pylint: disable=invalid-name
"""
Useful basic routines for managing polarization using the Stokes/Mueller calculus

Todo:
    * complete Jupyter notebook documentation
    * test mueller_to_jones()
    * improve internal documentation

Scott Prahl
May 2018
"""

import numpy as np
import pypolar.jones
import pypolar.fresnel

__all__ = ['op_linear_polarizer',
           'op_retarder',
           'op_attenuator',
           'op_mirror',
           'op_rotation',
           'op_quarter_wave_plate',
           'op_half_wave_plate',
           'op_fresnel_reflection',
           'op_fresnel_transmission',
           'stokes_linear',
           'stokes_left_circular',
           'stokes_right_circular',
           'stokes_horizontal',
           'stokes_vertical',
           'stokes_unpolarized',
           'intensity',
           'degree_of_polarization',
           'ellipse_orientation',
           'ellipse_ellipticity',
           'ellipse_axes',
           'stokes_to_jones',
           'mueller_to_jones',
           'interpret']


def op_linear_polarizer(theta):
    """
    Mueller matrix operator for a linear polarizer rotated about a normal to
    the surface of the polarizer.

    theta: rotation angle measured from the horizontal plane [radians]
    """

    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    lp = np.array([[1, C2, S2, 0],
                   [C2, C2**2, C2 * S2, 0],
                   [S2, C2 * S2, S2 * S2, 0],
                   [0, 0, 0, 0]])
    return 0.5 * lp


def op_retarder(theta, delta):
    """
    Mueller matrix operator for an optical retarder rotated about a normal to
    the surface of the retarder.

    theta: rotation angle between fast-axis and the horizontal plane [radians]
    delta: phase delay introduced between fast and slow-axes         [radians]
    """

    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    C = np.cos(delta)
    S = np.sin(delta)
    ret = np.array([[1, 0, 0, 0],
                    [0, C2**2 + C * S2**2, (1 - C) * S2 * C2, -S * S2],
                    [0, (1 - C) * C2 * S2, S2**2 + C * C2**2, S * C2],
                    [0, S * S2, -S * C2, C]])
    return ret


def op_attenuator(transmittance):
    """
    Mueller matrix operator for an optical attenuator.

    t : fraction of light getting through attenuator [---]
    """
    att = np.array([[t, 0, 0, 0],
                    [0, t, 0, 0],
                    [0, 0, t, 0],
                    [0, 0, 0, t]])
    return att


def op_mirror():
    """
    Mueller matrix operator for a perfect mirror.
    """

    mir = np.array([[1, 0, 0, 0],
                    [0, 1, 0, 0],
                    [0, 0, -1, 0],
                    [0, 0, 0, -1]])
    return mir


def op_rotation(theta):
    """
    Mueller matrix operator to rotate light about the optical axis.

    theta: rotation angle  [radians]
    """

    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    rot = np.array([[1, 0, 0, 0],
                    [0, C2, S2, 0],
                    [0, -S2, C2, 0],
                    [0, 0, 0, 1]])
    return rot


def op_quarter_wave_plate(theta):
    """
    Mueller matrix operator for an quarter-wave plate rotated about a normal to
    the surface of the plate.

    theta: rotation angle between fast-axis and the horizontal plane [radians]
    """

    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    qwp = np.array([[1, 0, 0, 0],
                    [0, C2**2, C2 * S2, -S2],
                    [0, C2 * S2, S2 * S2, C2],
                    [0, S2, -C2, 0]])
    return qwp


def op_half_wave_plate(theta):
    """
    Mueller matrix operator for an half-wave plate rotated about a normal to
    the surface of the plate.

    theta: rotation angle between fast-axis and the horizontal plane [radians]
    """

    C2 = np.cos(2 * theta)
    S2 = np.sin(2 * theta)
    qwp = np.array([[1, 0, 0, 0],
                    [0, C2**2 - S2**2, 2 * C2 * S2, 0],
                    [0, 2 * C2 * S2, S2 * S2 - C2**2, 0],
                    [0, 0, 0, -1]])
    return qwp


def op_fresnel_reflection(m, theta):
    """
    Mueller matrix operator for Fresnel reflection at angle theta

    Convert from the Jones operator to ensure that phase change are
    handled properly
    Args:
        m :     complex index of refraction   [-]
        theta : angle from normal to surface  [radians]
    Returns:
        4x4 Fresnel reflection operator       [-]
    """
    J = pypolar.jones.op_fresnel_reflection(m, theta)
    R = pypolar.jones.jones_op_to_mueller_op(J)
    return R


def op_fresnel_transmission(m, theta):
    """
    Mueller matrix operator for Fresnel transmission at angle theta

    Unclear if phase changes are handled properly.  See Collett, "Mueller-Stokes
    Matrix Formulation of Fresnel's Equations," Am. J. Phys. 39, 517 (1971).

    Args:
        m :     complex index of refraction       [-]
        theta : angle from normal to surface      [radians]
    Returns:
        4x4 Fresnel transmission operator         [-]
    """
    tau_p = pypolar.fresnel.T_par(m, theta)
    tau_s = pypolar.fresnel.T_per(m, theta)
    a = tau_s + tau_p
    b = tau_s - tau_p
    c = 2 * np.sqrt(tau_s*tau_p)
    mat = np.array([[a, b, 0, 0],
                    [b, a, 0, 0],
                    [0, 0, c, 0],
                    [0, 0, 0, c]])
    return 0.5 * mat


def stokes_linear(theta):
    """Stokes vector for linear polarized light at angle theta from
       horizontal plane"""

    return np.array([1, np.cos(2*theta), np.sin(2*theta), 0])


def stokes_right_circular():
    """Stokes vector corresponding to right circular polarized light"""

    return np.array([1, 0, 0, 1])


def stokes_left_circular():
    """Stokes vector corresponding to left circular polarized light"""

    return np.array([1, 0, 0, -1])


def stokes_horizontal():
    """
    Stokes vector corresponding to horizontal polarized light
    """
    return np.array([1, 1, 0, 0])


def stokes_vertical():
    """
    Stokes vector corresponding to vertical polarized light
    """
    return np.array([1, -1, 0, 0])


def stokes_unpolarized():
    """
    Stokes vector corresponding to vertical polarized light
    """
    return np.array([1, 0, 0, 0])


def intensity(S):
    """
    Returns the intensity
    """
    return S[0]


def degree_of_polarization(S):
    """
    Returns the degree of polarization
    """
    return S[0]/np.sqrt(S[1]**2+s[2]**2+s[3]**2)


def ellipse_orientation(S):
    """
    Returns the angle between the major semi-axis and the x-axis of
    the polarization ellipse (often represented by psi)
    """
    return 1/2 * np.arctan2(S[2], S[1])


def ellipse_ellipticity(S):
    """
    Returns the ellipticity of the polarization ellipse (often represented by chi)
    """
    return 1/2 * np.arcsin(S[3]/S[0])


def ellipse_axes(S):
    """
    Returns the semi-major and semi-minor axes of the polarization ellipse.
    """
    absL = np.sqrt(S[1]**2 + S[2]**2)
    A = np.sqrt((S[0] + absL)/2)
    B = np.sqrt((S[0] - absL)/2)
    return A, B


def stokes_to_jones(S):
    """
    Convert a Stokes vector to a Jones vector

    The Jones vector only has the polarized part of the intensity, of course.
    Also, since the Jones vector can differ by an arbitrary phase, the phase
    is chosen which makes the horizontal component purely real.

    Inputs:
        S : a Stokes vector

    Returns:
         the Jones vector corresponding to
    """

    # Calculate the degree of polarization
    p = np.sqrt(S[1]**2 + S[2]**2 + S[3]**2) / S[0]

    # Normalize the Stokes parameters (first one will be 1, of course)
    Q = S[1] / (S[0] * p)
    U = S[2] / (S[0] * p)
    V = S[3] / (S[0] * p)

    # the Jones components
    A = np.sqrt((1 + Q) / 2)
    if A == 0:
        B = 1
    else:
        B = complex(U, -V) / (2 * A)

    # put them together in a vector with the amplitude of the polarized part
    return np.sqrt(S[0] * p) * np.array([A, B])


def mueller_to_jones(M):
    """
    Convert a Mueller matrix to a Jones matrix

    Theocaris, Matrix Theory of Photoelasticity, eqns 4.70-4.76, 1979

    Inputs:
        M : a 4x4 Mueller matrix

    Returns:
         the corresponding 2x2 Jones matrix
    """
    A = np.empty((2, 2))
    A[0, 0] = np.sqrt((M[0, 0]+M[0, 1]+M[1, 0]+M[1, 1])/2)
    A[0, 1] = np.sqrt((M[0, 0]+M[0, 1]-M[1, 0]-M[1, 1])/2)
    A[1, 0] = np.sqrt((M[0, 0]-M[0, 1]+M[1, 0]-M[1, 1])/2)
    A[1, 1] = np.sqrt((M[0, 0]-M[0, 1]-M[1, 0]+M[1, 1])/2)

    theta = np.empty((2, 2))
    theta[0, 0] = 0
    theta[0, 1] = -np.arctan2(M[0, 3]+M[1, 3], M[0, 2]+M[1, 2])
    theta[1, 0] = np.arctan2(M[3, 0]+M[3, 1], M[2, 0]+M[2, 1])
    theta[1, 1] = np.arctan2(M[3, 2]-M[2, 3], M[2, 2]+M[3, 3])

    return A*np.exp(1j*theta)


def interpret(S):
    '''
    Interprets a Stokes vector

    Parameters
    S    : A Stokes vector

    Examples
    -------
    interpret([1, 0, 0, 0]) --> "Unpolarized Light"

    '''

    try:
        S0, S1, S2, S3 = S
    except:
        print("Stokes vector must have four real elements")
        return 0

    eps = 1e-12
    s = "not implemented yet"
    return s
